/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

/*! \file ors.h
    \brief Robot Simulation tools */

#ifndef MT_ors_h
#define MT_ors_h

#include "array.h"
#include "util.h"


//===========================================================================

// body types
#define BNONE -1
#define BBOX 0
#define BSPHERE 1
#define BCCYLINDER 2
#define BMESH 3
#define BCYLINDER 4

// joint types
#define JHINGE 0
#define JUNIVERSAL 1
#define JFIXED 2
#define JSLIDER 3
#define JBALL 4
#define JGLUE 5



namespace ors{
  //===========================================================================
  //! a 3D vector (double[3])
  struct Vector{
    double v[3];
    
    Vector(){}
    Vector(double x,double y,double z){ set(x,y,z); }
    double& operator()(int);
    const double& operator()(int) const;
    
    void set(double,double,double);
    void set(double*);
    void setZero();
    void setRandom();
    void add     (double,double,double);
    void subtract(double,double,double);
    void normalize();
    void setLength(double);
    void makeNormal  (const Vector&);
    void makeColinear(const Vector&);
    
    bool   isZero() const;
    bool   isNormalized() const;
    double isColinear(const Vector&) const;
    double length() const;
    double lengthSqr() const;
    double angle(const Vector&) const;
    double radius() const;
    double phi() const;
    double theta() const;
    
    void write(std::ostream&) const;
    void read (std::istream&);
  };
  
  //===========================================================================
  //! a matrix in 3D (double[9])
  struct Matrix{
    double m[9];

    Matrix(){};
    double& operator()(int,int);
    const double& operator()(int,int) const;
    
    void setZero();
    void setId();
    void setFrame   (Vector&,Vector&,Vector&);
    void setInvFrame(Vector&,Vector&,Vector&);
    void setXrot(double);
    void setSkew(const Vector&);
    void setExponential(const Vector&);
    void setOdeMatrix(double*);
    void setTensorProduct(const Vector&,const Vector&);

    void write(std::ostream&) const;
    void read (std::istream&);
  };
  
  //===========================================================================
  //! a quaterion (double[4])
  struct Quaternion{
    double q[4];
    
    Quaternion();
    
    void set(double q0,double x,double y,double z);
    void set(double* q);
    void setZero();
    void setRandom();
    void setDeg(double degree ,double axis0,double axis1,double axis2);
    void setDeg(double degree ,const Vector& axis);
    void setRad(double radians,double axis0,double axis1,double axis2);
    void setRad(double radians,const Vector& axis);
    void setRad(double angle);
    void setRadX(double angle);
    void setRadY(double angle);
    void setQuat(double s,double x,double y,double z);
    void setVec(Vector w);
    void setMatrix(double* m);
    void setDiff(const Vector& from,const Vector& to);
    void setInterpolate(double t,const Quaternion& a,const Quaternion b);
    void invert();
    void normalize();
    void multiply(double f);
    void alignWith(const Vector& v);
    
    bool isZero() const;
    bool isNormalized() const;
    double  getDeg() const;
    double  getRad() const;
    void    getDeg(double& degree,Vector& axis) const;
    void    getRad(double& angle ,Vector& axis) const;
    Vector& getVec(Vector& v) const;
    Vector& getX  (Vector& Rx) const;
    Vector& getY  (Vector& Ry) const;
    Vector& getZ  (Vector& Rz) const;
    double* getMatrix(double* m) const;
    double* getMatrixOde(double* m) const; //in Ode foramt: 3x4 memory storae
    double* getMatrixGL(double* m) const; //in OpenGL format: transposed 4x4 memory storage
    
    void writeNice(std::ostream& os) const;
    void write(std::ostream& os) const;
    void read(std::istream& is);
  };

  //===========================================================================
  //! a coordinate frame in 3D (position, orientation, linear & angular velocities)
  struct Frame{
    Vector p;  //!< position
    Quaternion r;//!< orientation (rotation w.r.t. world frame)
    Vector v;  //!< linear velocity
    Vector w;  //!< angular velocity
    
    Frame();

    void setZero();
    Frame& setText(const char* txt);
    void setRandom();
    void setInverse(const Frame& f);
    void setDifference(const Frame& from,const Frame& to);
    
    void addRelativeTranslation   (double x,double y,double z);
    void addRelativeVelocity      (double x,double y,double z);
    void addRelativeAngVelocityDeg(double degree,double x,double y,double z);
    void addRelativeAngVelocityRad(double rad,double x,double y,double z);
    void addRelativeAngVelocityRad(double wx,double wy,double wz);
    void addRelativeRotationDeg   (double degree,double x,double y,double z);
    void addRelativeRotationRad   (double rad,double x,double y,double z);
    void addRelativeRotationQuat  (double s,double x,double y,double z);
    
    void addRelativeFrame(const Frame& f);     // this = this * f
    void subRelativeFrame(const Frame& f);     // this = this * f^{-1}
    void makeRelative(const Frame& f);         // this = f * this
    void makeRelativeToInv(const Frame& f);    // this = f^{-1} * this
    
    double* getAffineMatrix(double *m) const;         // 4x4 matrix with 3x3=rotation and right-column=translation
    double* getInverseAffineMatrix(double *m) const;  // 4x4 matrix with 3x3=R^{-1}   and bottom-row=R^{-1}*translation
    double* getAffineMatrixGL(double *m) const;       // in OpenGL format (transposed memory storage!!)
    double* getInverseAffineMatrixGL(double *m) const;// in OpenGL format (transposed memory storage!!)
    
    void write(std::ostream& os) const;
    void read(std::istream& is);
  };

  //===========================================================================
  //! a mesh (arrays of vertices, triangles, colors & normals)
  struct Mesh{
    arr V;                //!< vertices
    arr Vn;               //!< triangle normals
    arr C;                //!< vertex colors
    intA G;               //!< vertex groups

    uintA T;              //!< triangles (faces)
    arr   Tn;             //!< vertex normals
    MT::Array<Frame*> GF; //!< frame for each group (GF.N is number of groups)
    MT::Array<uintA>  GT; //!< triangles for each group (GT.N=GF.N+1, last entry contains mixed group triangles)
    //MT::Array<uintA> strips; //!< triangle strips (each with a 1D stripe index set)
    
    Mesh();

    //set or create
    void clear();
    void setBox();
    void setTetrahedron();
    void setOctahedron();
    void setDodecahedron();
    void setSphere(uint fineness=4);
    void setHalfSphere(uint fineness=4);
    void setCylinder(double r,double l,uint fineness=4);
    void setCappedCylinder(double r,double l,uint fineness=4);
    void setGrid(uint X,uint Y);
    void setImplicitSurface(double(*fct)(double,double,double),double lo,double hi,uint res);

    //transform and modify
    void subDevide();
    void scale(double f);
    void scale(double sx,double sy,double sz);
    void translate(double dx,double dy,double dz);
    void center();
    void box();
    void addMesh(const ors::Mesh& mesh2);

    //internal computations & cleanup
    void computeNormals();
    void deleteUnusedVertices();
    void fuseNearVertices(double tol=1e-5);
    void clean();
    void flipFaces();
    void makeVerticesRelativeToGroup();

    //[preliminary]]
    void collectTriGroups();
    void skin(uint i);


    //IO
    void readFile(const char* filename);
    void readTriFile(const char* filename);
    void readObjFile(const char* filename);
    void readOffFile(const char* filename);
    void readPlyFile(const char* filename);
    void readStlFile(const char* filename);
    void writeTriFile(const char* filename);
    void writeOffFile(const char* filename);
  };

  //===========================================================================
  //! a spline
  struct Spline{
    uint T,K,degree;
    arr points;
    arr times;
    arr weights;
    arr basis,basis_trans,basis_timeGradient;
    
    void plotBasis();
    void setBasis();
    void setBasisAndTimeGradient();
    void setUniformNonperiodicBasis(uint T,uint K,uint degree);
    void evalF(arr& f_t,uint t) const;
    void evalF(arr& f) const;
    
    void partial(arr& dCdx,const arr& dCdf) const;
    void partial(arr& dCdx,arr& dCdt,const arr& dCdf,bool constrain=true) const;
  };

  //===========================================================================
  struct Joint;
  struct Shape;
  //! a rigid body (inertia properties, lists of attached joints & shapes)
  struct Body{
    uint index;          //!< unique identifier
    MT::Array<Joint*> inLinks,outLinks;       //!< lists of in and out joints

    MT::String name;     //!< name
    Frame X;             //!< body's absolute inertial frame
    AnyList ats;         //!< list of any-type attributes

    //dynamic properties
    bool fixed;          //!< is globally fixed?
    double mass;         //!< its mass
    Matrix inertia;      //!< its inertia tensor
    Vector cog;          //!< its center of gravity
    Vector force,torque; //!< current forces applying on the body
    
    MT::Array<Shape*> shapes;

    Body();
    ~Body();
    void operator=(const Body& b){
      index=b.index; name=b.name; X=b.X; listClone(ats,b.ats);
      fixed=b.fixed; mass=b.mass; inertia=b.inertia; cog=b.cog; force=b.force; torque=b.torque; }
    void reset();
    void write(std::ostream& os) const;
    void read(std::istream& is);
  };

  //===========================================================================
  //! a joint
  struct Joint{
    uint index;          //!< unique identifier
    int a,b;             //!< indices of from and to bodies
    Body *from,*to;      //!< pointers to from and to bodies

    int type;            //!< joint type
    Frame A;             //!< transformation from parent body to joint (attachment, usually static)
    Frame Q;             //!< transformation within the joint (usually dynamic)
    Frame B;             //!< transformation from joint to child body (attachment, usually static)
    Frame Xworld;        //!< joint frame in world coordinates (same as A*from->X)
    AnyList ats;         //!< list of any-type attributes

    Joint(){ reset(); }
    ~Joint(){ reset(); }
    void reset(){ listDelete(ats); A.setZero(); B.setZero(); Q.setZero(); Xworld.setZero(); p[0]=p[1]=0.; motor=fixed=false; type=JHINGE; }
    void write(std::ostream& os) const;
    void read(std::istream& is);
    Joint &data(){ return *this; }
    
    //old trash:
    double p[2];        //joint parameters (limits)
    bool motor;         //are there motors in the joint?
    bool fixed;         //is this a rigit joint? 15. Mar 06 (hh)
    int z;
  };

  //===========================================================================
  //! a shape (geometric shape like cylinder/mesh, associated to a body)
  struct Shape{
    MT::String name;     //!< name
    Frame X;
    Frame rel;      //!< relative translation/rotation of the bodies geometry
    int type;
    double size[4];
    double color[3];
    Mesh mesh;
    bool cont;      //!< are contacts registered (or filtered in the callback)
    Vector contactOrientation;

    uint index;
    Body *body;

    AnyList ats;    //!< list of any-type attributes

    Shape();
    ~Shape(){ reset(); }
    void operator=(const Shape& s){
      rel=s.rel; type=s.type;
      memmove(size,s.size,4*sizeof(double)); memmove(color,s.color,3*sizeof(double));
      mesh=s.mesh; cont=s.cont;  }
    void reset();
    void write(std::ostream& os) const;
    void read(std::istream& is);
  };
  
  //===========================================================================
  //! proximity information (when two shapes become close)
  struct Proxy{
    int a;             //!< index of body A
    int b;             //!< index of body B
    Vector posA,velA;  //!< contact or closest point position on surface of body A (in world coordinates)
    Vector posB,velB;  //!< contact or closest point position on surface of body B (in world coordinates)
    Vector normal;     //!< contact normal, pointing from A to B
    double d;          //!< distance (positive) or penetration (negative) between A and B
    Frame rel;         //!< relative frame from A to B WHEN the two bodies collided for the first time
    uint age;
  };

  //===========================================================================
  //! data structure to store a whole physical situation (lists of bodies, joints, shapes, proxies)
  struct Graph{
    typedef ors::Body* node;
    typedef ors::Joint* edge;

    //!@name data fields
    uint sd,jd,td; // added jd (joint dim.), 16. Mar 06 (hh)
    MT::Array<ors::Body*>  bodies;
    MT::Array<ors::Joint*> joints;
    MT::Array<ors::Shape*> shapes;
    MT::Array<ors::Proxy*> proxies; //!< list of current proximities between bodies

    //!@name constructors
    Graph(){ sd=jd=0; bodies.memMove=joints.memMove=shapes.memMove=proxies.memMove=true; }
    ~Graph(){ clear(); }
  
    //!@name initializations
    void init(const char* filename);
    
    //!@name changes of configuration
    void clear();
    void revertEdge(edge e);
    void reconfigureRoot(const node& n);
    void transformEdge(edge e,const ors::Frame &f); //A <- A*f, B <- f^{-1}*B
    void zeroGaugeEdges();                            //A <- A*Q, Q <- Id
    void glueBodies(node a,node b);
    void glueTouchingBodies();

    //!@name computations on the DoFs
    void calcNodeFramesFromEdges();
    void calcEdgesFromNodeFrames();
    void clearEdgeErrors();
    void invertTime();
    void computeNaturalQmetric(arr& W);

    //!@name kinematics & dynamics
    void kinematics(arr& x,uint i,ors::Frame *rel=0);
    void jacobian(arr& J,uint i,ors::Frame *rel=0);
    void hessian(arr& H,uint i,ors::Frame *rel=0);
    void inertia(arr& M);
    void equationOfMotion(arr& M,arr& F,const arr& qd);
    void dynamics(arr& qdd,const arr& qd,const arr& tau);
    void inverseDynamics(arr& tau,const arr& qd,const arr& qdd);
    void jacobianAll(arr& J);
    void kinematicsZ(arr& z,uint i,ors::Frame *rel=0);
    void jacobianZ  (arr& J,uint i,ors::Frame *rel=0);
    void jacobianR  (arr& J,uint i);

    //!@name get state
    uint getJointStateDimension() const;
    void getJointState(arr& x,arr& v);
    void getJointState(arr& x);
    uint getFullStateDimension();
    void getFullState(arr& x);
    void getFullState(arr& x,arr& v);
    void getContactMeasure(arr &x,double margin=.02);
    double getContactGradient(arr &grad,double margin=.02);
    void getLimitsMeasure(arr &x,const arr& limits,double margin=.1);
    double getLimitsGradient(arr &grad,const arr& limits,double margin=.1);
    void getComGradient(arr &grad);
    void getTotals(ors::Vector& c,ors::Vector& v,ors::Vector& l,ors::Quaternion& ori);
    void getGyroscope(ors::Vector& up);
    double getEnergy();
    double getCenterOfMass(arr& com);
    double getJointErrors();
    void getPenetrationState(arr &vec);
    uint getTouchDimension();
    void getTouchState(arr& touch);  
    void getGripState(arr& grip,uint j);
    ors::Proxy* getContact(uint a,uint b);

    //!@name set state
    void setJointState(const arr& x,const arr& v,bool clearJointErrors=true);
    void setJointState(const arr& x,bool clearJointErrors=true);
    void setFullState(const arr& x,bool clearJointErrors=true);
    void setFullState(const arr& x,const arr& v,bool clearJointErrors=true);
    void clearForces();
    void addForce(ors::Vector force,node n,ors::Vector pos);
    void contactsToForces(double hook=.01,double damp=.0003);
    void gravityToForces();
    void frictionToForces(double coeff);

    //!@name I/O
    void reportProxies(std::ostream *os=&std::cout);
    void reportGlue(std::ostream *os=&std::cout);

    void sortProxies(bool deleteMultiple=false,bool deleteOld=false);

    bool checkUniqueNames() const;
    node getName(const char* name){ return getBodyByName(name); } //obsolete
    Body  *getBodyByName(const char* name);
    Shape *getShapeByName(const char* name);
    edge getName(const char* from,const char* to);
    void prefixNames();

    void write(std::ostream& os) const;
    void read(std::istream& is);
  };
}


namespace ors{
  double  operator*(const Vector&,const Vector&);
  Vector  operator^(const Vector&,const Vector&);
  Vector  operator+(const Vector&,const Vector&);
  Vector  operator-(const Vector&,const Vector&);
  Vector  operator*(double,const Vector&);
  Vector  operator*(const Vector&,double);
  Vector  operator/(const Vector&,double);
  Vector& operator*=(Vector&,double);
  Vector& operator/=(Vector&,double);
  Vector& operator+=(Vector&,const Vector&);
  Vector& operator-=(Vector&,const Vector&);
  Vector  operator-(const Vector&);

  Matrix  operator*(const Matrix& b,const Matrix& c);
  Matrix  operator+(const Matrix& b,const Matrix& c);
  Vector  operator*(const Matrix& b,const Vector& c);
  Matrix& operator*=(Matrix& a,double c);
  Matrix  operator*(double b,const Matrix& c);
  Matrix& operator+=(Matrix& a,const Matrix& b);

  Quaternion operator*(const Quaternion& b,const Quaternion& c);
  Quaternion operator/(const Quaternion& b,const Quaternion& c);
  Vector operator*(const Quaternion& b,const Vector& c);
  Vector operator/(const Quaternion& b,const Vector& c);
  Vector operator*(const Frame& b,const Vector& c);
  Vector operator/(const Frame& b,const Vector& c);
}
std::istream& operator>>(std::istream&,ors::Vector&);
std::istream& operator>>(std::istream&,ors::Matrix&);
std::istream& operator>>(std::istream&,ors::Quaternion&);
std::istream& operator>>(std::istream&,ors::Frame&);
std::istream& operator>>(std::istream&,ors::Body&);
std::istream& operator>>(std::istream&,ors::Joint&);
std::istream& operator>>(std::istream&,ors::Proxy&);
std::ostream& operator<<(std::ostream&,const ors::Vector&);
std::ostream& operator<<(std::ostream&,const ors::Matrix&);
std::ostream& operator<<(std::ostream&,const ors::Quaternion&);
std::ostream& operator<<(std::ostream&,const ors::Frame&);
std::ostream& operator<<(std::ostream&,const ors::Body&);
std::ostream& operator<<(std::ostream&,const ors::Joint&);
std::ostream& operator<<(std::ostream&,const ors::Proxy&);
stdPipes(ors::Graph);

double scalarProduct(const ors::Quaternion& a,const ors::Quaternion& b);



//===========================================================================
//
// C-style functions
//

void inertiaSphere  (double *Inertia, double& mass, double density, double radius);
void inertiaBox     (double *Inertia, double& mass, double density, double dx, double dy, double dz);
void inertiaCylinder(double *Inertia, double& mass, double density, double height, double radius);



//===========================================================================
//
// constants
//

extern const ors::Vector VEC_x;
extern const ors::Vector VEC_y;
extern const ors::Vector VEC_z;


//===========================================================================
//
// fileIO
//

void readBlender(const char* filename,ors::Mesh& mesh,ors::Graph& bl);


#ifdef MT_IMPLEMENTATION
#include "ors.cpp"
//#include "ors_blender.cpp"
#endif


#endif
