/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "ors_swift.h"
#include "ors_control.h"
#include "ors_opengl.h"
#include "opengl.h"
#include "soc.h"

//===========================================================================
//
// OrsSocImplementation
//

struct soc::OrsSocWorkspace{
  ors::Graph *ors;
  SwiftModule *swift;
  ControlVariableList vars;
  
  arr Qlin,Qoff,Qinv;
  
  uint T;
  arr q0,v0,W,H,Q,v_act;
  double tau;
  bool pseudoDynamic;
};

soc::OrsSocImplementation::OrsSocImplementation(){
  WS = new OrsSocWorkspace;
  WS->ors   = NULL; //new ors::Graph;
  WS->swift = NULL; //new SwiftModule;
  WS->vars = globalSpace;
  WS->pseudoDynamic=false;
}

soc::OrsSocImplementation::~OrsSocImplementation(){
  delete WS;
}

void soc::OrsSocImplementation::initKinematic(ors::Graph *ors,SwiftModule *swift,OpenGL *_gl,uint trajectory_length, arr*W){
  WS->ors   = ors;
  WS->swift = swift;
  gl        = _gl;
  WS->T=trajectory_length;
  WS->tau=.01;
  WS->ors->getJointState(WS->q0,WS->v0);
  if(W){
    if(W->nd==1){ CHECK(W->N==WS->q0.N,""); WS->W.setDiag(*W); }
    else NIY;
  }else{
    WS->ors->computeNaturalQmetric(WS->W);
  }
  WS->H = WS->W;
  dynamic=false;
}

void soc::OrsSocImplementation::initPseudoDynamic(ors::Graph *ors,SwiftModule *swift,OpenGL *_gl, double trajectory_time,  uint trajectory_steps, arr *W){
  WS->ors   = ors;
  WS->swift = swift;
  gl        = _gl;
  setTimeInterval(trajectory_time, trajectory_steps);
  setq0AsCurrent();
  if(W){
    if(W->nd==1){ CHECK(W->N==WS->q0.N,""); WS->W.setDiag(*W); }
    else NIY;
  }else{
    WS->ors->computeNaturalQmetric(WS->W);
  }
  static MT::Parameter<double> hc("Hcost");
  static MT::Parameter<double> qn("Qnoise",1e-10);
  uint n=WS->q0.N;
  WS->H = hc()*WS->W;  //u-metric for torque control
  WS->Q.setDiag(qn,2*n);  //covariance \dot q-update
  dynamic=true;
  WS->pseudoDynamic=true;
}

void soc::OrsSocImplementation::setTimeInterval(double trajectory_time,  uint trajectory_steps){
  WS->T=trajectory_steps;
  WS->tau=trajectory_time/trajectory_steps;
}

void soc::OrsSocImplementation::setControlVariables(const ControlVariableList& _CVlist){
  WS->vars=_CVlist;
}

//! report on some 
void soc::OrsSocImplementation::reportOnState(ostream& os){
  os <<"OrsSocImplementat - state report:\n";
  os <<"** control variables:" <<endl;
  reportAll(WS->vars,os);
  os <<"** proxies:" <<endl;
  WS->ors->reportProxies(&os);
}


uint soc::OrsSocImplementation::nTime(){  return WS->T; }
uint soc::OrsSocImplementation::nTasks(){ return WS->vars.N; }
uint soc::OrsSocImplementation::qDim(){   if(!WS->Qlin.N) return WS->q0.N; else return WS->Qlin.d1; }
uint soc::OrsSocImplementation::uDim(){   return qDim(); }
uint soc::OrsSocImplementation::yDim(uint i){ return WS->vars(i)->x.N; }
double soc::OrsSocImplementation::getTau(){  return WS->tau; }

void soc::OrsSocImplementation::getq0 (arr& q) { if(!WS->Qlin.N) q=WS->q0; else q=WS->Qinv*(WS->q0-WS->Qoff); }
void soc::OrsSocImplementation::getv0 (arr& v) { if(!WS->Qlin.N) v=WS->v0; else v=WS->Qinv*WS->v0; }
void soc::OrsSocImplementation::getqv0(arr& q_){
   if(!WS->Qlin.N) q_.setBlockVector(WS->q0,WS->v0);
   else{
     arr q,qd;
     getqv0(q,qd);
     q_.setBlockVector(q,qd);
   }
}
void soc::OrsSocImplementation::getqv0(arr& q,arr& qd){ getq0(q); getv0(qd); }
void soc::OrsSocImplementation::getW  (arr& W) { if(!WS->Qlin.N) W=WS->W; else NIY; }
void soc::OrsSocImplementation::getH  (arr& H) { if(!WS->Qlin.N) H=WS->H; else H = ~WS->Qlin*WS->H*WS->Qlin; }
void soc::OrsSocImplementation::getQ  (arr& Q) {
  if(!WS->Qlin.N) Q=WS->Q; else{
    arr Qbig;
    Qbig.resize(2*WS->Qlin.d0,2*WS->Qlin.d1); Qbig.setZero();
    Qbig.setMatrixBlock(WS->Qlin,0,0); Qbig.setMatrixBlock(WS->Qlin,WS->Qlin.d0,WS->Qlin.d1);
    //cout <<Qbig <<endl;
    Q = ~Qbig*WS->Q*Qbig;
  }
}

void soc::OrsSocImplementation::setq(const arr& _q){
  arr q;
  if(!WS->Qlin.N) q.referTo(_q);
  else q = WS->Qlin*_q + WS->Qoff;
  WS->ors->setJointState(q);
  WS->ors->calcNodeFramesFromEdges();
  WS->swift->computeProxies(*WS->ors,false);
  WS->v_act.resizeAs(q);
  WS->v_act.setZero();
  uint i;
  ControlVariable *v;
  for_list(i,v,WS->vars) if(v->active){
    v->updateState();
    v->updateJacobian();
  }
}

void soc::OrsSocImplementation::setqv(const arr& _q,const arr& _qd){
  arr q,qd;
  if(!WS->Qlin.N){ q.referTo(_q); qd.referTo(_qd); }
  else{ q = WS->Qlin*_q + WS->Qoff;  qd = WS->Qlin*_qd; }
  WS->ors->setJointState(q,qd);
  WS->ors->calcNodeFramesFromEdges();
  WS->swift->computeProxies(*WS->ors,false);
  WS->v_act=qd;
  uint i;
  ControlVariable *v;
  for_list(i,v,WS->vars) if(v->active){
    v->updateState();
    v->updateJacobian();
  }
}

void soc::OrsSocImplementation::setqv(const arr& q_){
  uint n=q_.N/2;
  CHECK(q_.N==2*n,"");
  arr q,v;
  q.referToSubRange(q_,0,n-1);
  v.referToSubRange(q_,n,2*n-1);
  setqv(q,v);
}
void soc::OrsSocImplementation::setq0AsCurrent(){
  WS->ors->getJointState(WS->q0,WS->v0);
}


void soc::OrsSocImplementation::getMF(arr& M,arr& F){
  if(!WS->pseudoDynamic){
    if(WS->Qlin.N) NIY;
    WS->ors->clearForces();
    WS->ors->gravityToForces();
    //WS->ors->frictionToForces(1.1);
    WS->ors->equationOfMotion(M,F,WS->v_act);
    //M.setId();  F = .1;
    //Minv *= .2;//1e-1;
  }else{
    uint n=qDim();
    M.setId(n);
    F.resize(n); F.setZero();
  }
}

void soc::OrsSocImplementation::getMinvF(arr& Minv,arr& F){
  arr M;
  getMF(M,F);
  inverse(Minv,M);
}

bool soc::OrsSocImplementation::isConditioned(uint i,uint t){
  return WS->vars(i)->active;
}

const char* soc::OrsSocImplementation::taskName(uint i){
  return WS->vars(i)->name.p;
}

void soc::OrsSocImplementation::getPhi(arr& phiq_i,uint i){
  //WS->vars(i)->updateState();
  phiq_i=WS->vars(i)->x;
}

void soc::OrsSocImplementation::getJqd(arr& jqd_i,uint i){
  arr q,qd;
  WS->ors->getJointState(q,qd);
  //WS->vars(i)->updateJacobian();
  jqd_i = WS->vars(i)->J * qd;
}

void soc::OrsSocImplementation::getJtJ(arr& J_i,arr& tJ_i,uint i){
  //WS->vars(i)->updateJacobian();
  if(!WS->Qlin.N){
    J_i=WS->vars(i)->J;
    tJ_i=WS->vars(i)->tJ;
  }else{
    J_i=WS->vars(i)->J*WS->Qlin;
    transpose(tJ_i,J_i);
  }
}

void soc::OrsSocImplementation::getHessian(arr& H_i,uint i){
  if(WS->Qlin.N) NIY;
  WS->vars(i)->getHessian(H_i);
}

void soc::OrsSocImplementation::getTarget(arr& y_i,double& prec,uint i,uint t){
  if(!t && WS->vars(i)->targetType!=trajectoryTT){
    ControlVariable *v=WS->vars(i);
    v->updateChange(-1);
    y_i  = v->x+v->x_change;
    prec = WS->vars(i)->prec;
    return;
  }
  y_i  = WS->vars(i)->x_trajectory[t];
  prec = WS->vars(i)->prec_trajectory(t);
}

void soc::OrsSocImplementation::getTargetV(arr& v_i,double& prec,uint i,uint t){
  if(!t && WS->vars(i)->targetType!=trajectoryTT){
    v_i  = WS->vars(i)->v_target;
    prec = WS->vars(i)->vprec;
    return;
  }
  v_i  = WS->vars(i)->v_trajectory[t];
  prec = WS->vars(i)->vprec_trajectory(t);
}


//===========================================================================
//
// problem implementations
//

void drawOrsSocEnv(void*){
  glStandardLight();
  //glDrawFloor(1.,.4,.4,.4);
  //DrawAxes(1.);
}

void soc::setupOpenGL(OrsSocImplementation &soci){
  if(!soci.gl) soci.gl=new OpenGL();
  soci.gl->add(drawOrsSocEnv,0);
  soci.gl->add(ors::glDrawGraph,soci.WS->ors);
  //soci.gl->add(plotDrawOpenGL,&plotData);
  soci.gl->camera.focus(0,0,.8);
}

void soc::createEndeffectorReachProblem(OrsSocImplementation &soci,
                                   const char *ors_file,
                                   uint trajectory_length,
                                   int rand_seed){

  //setup the workspace
  MT::load(*soci.WS->ors,ors_file);
  if(rand_seed>0){
    rnd.seed(rand_seed);
    ors::Body &t=*soci.WS->ors->getName("target");
    t.X.p(0) += .05*rnd.gauss();
    t.X.p(1) += .05*rnd.gauss();
    t.X.p(2) += .05*rnd.gauss();
  }
  soci.WS->ors->calcNodeFramesFromEdges();
  soci.WS->ors->reconfigureRoot(soci.WS->ors->getName("rfoot"));
  soci.WS->ors->getJointState(soci.WS->q0,soci.WS->v0);
  soci.WS->swift->init(*soci.WS->ors);


  soci.WS->T=trajectory_length;
  soci.WS->tau=.01;
  arr Wdiag(soci.WS->q0.N);
  Wdiag=1.;
  MT_MSG("Warning - need to change this");
  Wdiag <<"[20 20 20 10 10 10 10 1 1 1 1 10 10 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 20 20 10 10 10 10 10 10 ]";
  //Wdiag <<"[20 20 20 10 10 10 10 1 1 1 1 10 10 1 1 1 20 20 10 10 10 10 10 10 ]";
  CHECK(Wdiag.N==soci.WS->q0.N,"wrong W matrix!");
  soci.WS->W.setDiag(Wdiag);
  soci.WS->H = soci.WS->W;

  //set task variables
  ControlVariable *x0,*x1,*x2;
  x0 = new ControlVariable("finger-tip",*soci.WS->ors,posCVT ,"effector",0,0,0,0);
  x1 = new ControlVariable("COM",       *soci.WS->ors,comCVT ,0,0,0,0,0);
  x2 = new ControlVariable("collision", *soci.WS->ors,collCVT,0,0,0,0,ARR(.02));
  soci.WS->vars = TUPLE(x0,x1,x2);

  updateState(globalSpace);
  //reportAll(globalSpace,cout);


  double midPrec,endPrec,balPrec,colPrec;
  MT::getParameter(midPrec,"midPrec");
  MT::getParameter(endPrec,"endPrec");
  MT::getParameter(balPrec,"balPrec");
  MT::getParameter(colPrec,"colPrec");

  x0->x_target.setCarray(soci.WS->ors->getName("target")->X.p.v,3);
  x0->setInterpolatedTargetTrajectory(soci.WS->T);
  x0->setPrecisionTrajectoryFinal(soci.WS->T,midPrec,endPrec);

  x1->x_target(0)=0.;
  x1->x_target(1)=-.02;
  x1->setInterpolatedTargetTrajectory(soci.WS->T);
  x1->setPrecisionTrajectoryConstant(soci.WS->T,balPrec);
  if(!balPrec) x1->active=false;

  x2->x_target = 0.;
  x2->setInterpolatedTargetTrajectory(soci.WS->T);
  x2->setPrecisionTrajectoryConstant(soci.WS->T,colPrec);
  if(!colPrec) x2->active=false;

  soci.dynamic=false;
}

void soc::createDynamicProblem(OrsSocImplementation &soci,
                          const char *ors_file,
                          double trajectory_time,
                          uint trajectory_steps){

  //setup the workspace
  MT::load(*soci.WS->ors,ors_file);
  soci.WS->ors->calcNodeFramesFromEdges();
  soci.WS->ors->getJointState(soci.WS->q0,soci.WS->v0);
  soci.WS->swift->init(*soci.WS->ors);

  uint n=soci.WS->q0.N;
  soci.WS->T=trajectory_steps;
  soci.WS->tau=trajectory_time/trajectory_steps;
  soci.WS->W.setDiag(1e-6,n);  //q-metric for inverse kinematics (initialization)
  static MT::Parameter<double> hc("Hcost");
  static MT::Parameter<double> qn("Qnoise");
  soci.WS->H.setDiag(hc,n);  //u-metric for torque control
  soci.WS->Q.setDiag(qn,2*n);  //covariance \dot q-update

  //set task variables
  ControlVariable *x0;
  x0 = new ControlVariable("finger-tip",*soci.WS->ors,posCVT ,"eff","t(0 0 .15)",0,0,0);
  soci.WS->vars = TUPLE(x0);

  updateState(globalSpace);
  //reportAll(globalSpace,cout);

  double midPrec,endPrec;
  MT::getParameter(midPrec,"midPrec");
  MT::getParameter(endPrec,"endPrec");

  x0->x_target.setCarray(soci.WS->ors->getName("target")->X.p.v,3);
  x0->v_target <<"[2 0 0]";
  x0->setInterpolatedTargetTrajectory(soci.WS->T);
  x0->setPrecisionTrajectoryFinal (soci.WS->T,midPrec,endPrec);
  x0->setPrecisionVTrajectoryFinal(soci.WS->T,0.,endPrec);
  

  soci.dynamic=true;
}

void createNikolayReachProblem(soc::OrsSocImplementation &soci,
                                   ors::Graph &ors,
                                   SwiftModule& swift,
                                   uint trajectory_length,
                                   const arr& endeffector_target,
                                   const char* endeffector_name,
                                   const arr& W){
  static soc::OrsSocWorkspace WS;

  //setup the workspace
  WS.vars = globalSpace;
  WS.ors=&ors;
  WS.ors->getJointState(WS.q0,WS.v0);
  WS.T=trajectory_length;
  WS.W=W;
  WS.swift=&swift;

  //set task variables
  ControlVariable *x0,*x1;
  x0 = new ControlVariable("finger-tip",*WS.ors,posCVT ,endeffector_name,"",0,0,0);
  x1 = new ControlVariable("collision", *WS.ors,collCVT,0,0,0,0,0);
  WS.vars = TUPLE(x0,x1);

  updateState(globalSpace);
  //reportAll(globalSpace,cout);

  x0->prec=1e3;  x0->setGainsAsAttractor(30.);  x0->x_target = endeffector_target;
  x0->setTrajectory(WS.T,0,1e10);
  x0->setPrecisionTrajectoryFinal(WS.T,1e1,1e4);

  x1->prec=1e6;  x1->setGainsAsAttractor(30.);  x1->x_target = 0.;
  x1->setTrajectory(WS.T,0);
  x1->setPrecisionTrajectoryConstant(WS.T,1e5);

  soci.WS=&WS;
  soci.dynamic=false;
}
